#define BUILD_NUMBER \
	"build-1280544"
#define BUILD_NUMBER_NUMERIC \
	1280544
#define BUILD_NUMBER_NUMERIC_STRING \
	"1280544"
#define PRODUCT_BUILD_NUMBER \
	"product-build-25793"
#define PRODUCT_BUILD_NUMBER_NUMERIC \
	25793
#define PRODUCT_BUILD_NUMBER_NUMERIC_STRING \
	"25793"
